import itertools
from collections import defaultdict

import backtrader as bt


def get_plotobjs(strategy, include_non_plotable=False,
                 order_by_plotmaster=False):
    '''
    Returns all plotable objects of a strategy

    By default the result will be ordered by the
    data the object is aligned to. If order_by_plotmaster
    is True, objects will be aligned to their plotmaster.
    '''
    datas = strategy.datas
    inds = strategy.getindicators()
    obs = strategy.getobservers()
    objs = defaultdict(list)
    # ensure strategy is included
    objs[strategy] = []
    # first loop through datas
    for d in datas:
        if not include_non_plotable and not d.plotinfo.plot:
            continue
        objs[d] = []
    # next loop through all ind and obs and set them to
    # the corresponding data clock
    for obj in itertools.chain(inds, obs):
        # check for base classes
        if not isinstance(obj, (bt.IndicatorBase,
                                bt.MultiCoupler,
                                bt.ObserverBase)):
            continue
        # check for plotinfos
        if not hasattr(obj, 'plotinfo'):
            # no plotting support cause no plotinfo attribute
            # available - so far LineSingle derived classes
            continue
        # should this indicator be plotted?
        if (not include_non_plotable
                and (not obj.plotinfo.plot or obj.plotinfo.plotskip)):
            continue
        # append object to the data object
        pltmaster = get_plotmaster(obj)
        data = get_clock_obj(obj, True)
        if pltmaster in objs:
            objs[pltmaster].append(obj)
        elif data in objs:
            objs[data].append(obj)

    if not order_by_plotmaster:
        return objs

    # order objects by its plotmaster
    pobjs = defaultdict(list)
    for d in objs:
        pmaster = get_plotmaster(d)
        # add all datas, if a data has a plotmaster, add it to plotmaster
        if pmaster is d and pmaster not in pobjs:
            pobjs[pmaster] = []
        elif pmaster is not None and pmaster is not d:
            pobjs[pmaster].append(d)

        for o in objs[d]:
            pmaster = get_plotmaster(o.plotinfo.plotmaster)
            subplot = o.plotinfo.subplot
            if subplot and pmaster is None:
                pobjs[o] = []
            elif pmaster is not None:
                pobjs[pmaster].append(o)
            else:
                pmaster = get_plotmaster(get_clock_obj(o, True))
                if pmaster is not None and pmaster in pobjs:
                    pobjs[pmaster].append(o)
    # return objects ordered by plotmaster
    return pobjs


def get_plotmaster(obj):
    '''
    Resolves the plotmaster of the given object
    '''
    if obj is None:
        return None

    while True:
        pm = obj.plotinfo.plotmaster
        if pm is None:
            break
        else:
            obj = pm
    if isinstance(obj, bt.Strategy):
        return None
    return obj


def get_last_avail_idx(strategy, dataname=False):
    '''
    Returns the last available index of a data source
    '''
    if dataname is not False:
        data = strategy.getdatabyname(dataname)
    else:
        data = strategy
    offset = 0
    while True:
        if (len(data) > offset
                and data.datetime[-offset] != data.datetime[-offset]):
            offset += 1
            continue
        break
    return len(data) - 1 - offset


def filter_obj(obj, filter):
    '''
    Returns if the given object should be filtered.
    False if object should not be filtered out, 
    True if object should be filtered out.
    '''

    if filter is None:
        return False

    dataname = get_dataname(obj)
    plotid = obj.plotinfo.plotid

    # filter by dataname
    if 'dataname' in filter:
        if dataname is not False:
            if isinstance(filter['dataname'], str):
                if dataname != filter['dataname']:
                    return True
            elif isinstance(filter['dataname', list]):
                if dataname not in filter['dataname']:
                    return True
    if 'group' in filter:
        if isinstance(filter['group'], str):
            if filter['group'] != '':
                plotids = filter['group'].split(',')
                if plotid not in plotids:
                    return True

    return False


def get_datanames(strategy, filter=True):
    '''
    Returns the names of all data sources
    '''
    datanames = []
    for d in strategy.datas:
        if not filter or d.plotinfo.plot is not False:
            datanames.append(get_dataname(d))
    return datanames


def get_dataname(obj):
    '''
    Returns the name of the data for the given object
    If the data for a object is a strategy then False will
    be returned.
    '''
    data = get_clock_obj(obj, True)
    if isinstance(data, bt.Strategy):
        # strategy will have no dataname
        return False
    elif isinstance(data, bt.AbstractDataBase):
        # data feeds are end points
        # try some popular attributes that might carry a name
        # _name: user assigned value upon instantiation
        # _dataname: underlying bt dataname (is always available)
        # if that fails, use str
        for n in ['_name', '_dataname']:
            val = getattr(data, n)
            if val is not None:
                break
        if val is None:
            val = str(data)
        return val
    else:
        raise Exception(
            f'Unsupported data: {obj.__class__}')


def get_smallest_dataname(strategy, datanames):
    '''
    Returns the smallest dataname from a list of
    datanames
    '''
    data = False
    for d in datanames:
        if not d:
            continue
        tmp = strategy.getdatabyname(d)
        if (data is False
            or (tmp._timeframe < data._timeframe)
            or (tmp._timeframe == data._timeframe
                and tmp._compression < data._compression)):
            data = tmp
    if data is False:
        return data
    return get_dataname(data)


def get_clock_obj(obj, resolve_to_data=False):
    '''
    Returns a clock object to use for building data
    A clock object can be either a strategy, data source,
    indicator or a observer.
    '''
    if isinstance(obj, bt.LinesOperation):
        # indicators can be created to run on a line
        # (instead of e.g. a data object) in that case grab
        # the owner of that line to find the corresponding clock
        # also check for line actions like "macd > data[0]"
        return get_clock_obj(obj._clock, resolve_to_data)
    elif isinstance(obj, (bt.LineSingle)):
        # if we have a line, return its owners clock
        return get_clock_obj(obj._owner, resolve_to_data)
    elif isinstance(obj, bt.LineSeriesStub):
        # if its a LineSeriesStub object, take the first line
        # and get the clock from it
        return get_clock_obj(obj.lines[0], resolve_to_data)
    elif isinstance(obj, (bt.IndicatorBase, bt.MultiCoupler, bt.ObserverBase)):
        # a indicator and observer can be a clock, internally
        # it is obj._clock
        if resolve_to_data:
            return get_clock_obj(obj._clock, resolve_to_data)
        clk = obj
    elif isinstance(obj, bt.StrategyBase):
        # a strategy can be a clock, internally it is obj.data
        clk = obj
    elif isinstance(obj, bt.AbstractDataBase):
        clk = obj
    else:
        raise Exception(
            f'Unsupported clock: {obj.__class__}')
    return clk


def get_clock_line(obj):
    '''
    Find the corresponding clock for an object.
    A clock is a datetime line that holds timestamps
    for the line in question.
    '''
    clk = get_clock_obj(obj)
    return clk.lines.datetime


def get_source_id(source):
    '''
    Returns a unique source id for given source.
    This is used for unique column names.
    '''
    return str(id(source))



import asyncio
import logging
from threading import Lock
import threading

import backtrader as bt

import tornado.ioloop

from ..app import BacktraderBokeh
from ..webapp import Webapp
from ..schemes import Tradimo, Blackly
from ..live.client import LiveClient

_logger = logging.getLogger(__name__)

def combine_custom_conf(name, default):
    custom_config = dict(bt.custom_config['system'], **bt.custom_config['browser'])
    if name in custom_config:
        val = custom_config[name]
        if name == 'scheme':
            if val in ['White', 'Tradimo']:
                val = Tradimo()
            elif val in ['Black', 'Blackly']:
                val = Blackly()
            else:
                val = default

        if name == 'tabs':
            try:
                v = []
                for one in val:
                    if one == 'LogTabs':
                        col = bt.custom_config["scheme"]["logger_tab_num_cols"] if bt.custom_config["scheme"]["logger_tab_num_cols"] else 2
                        one = f'LogTabs({col})'
                    v.append(eval('bt.tabs.' + one))
                val = v
            except:
                val = default

        return val
    else:
        return default

class LivePlotAnalyzer(bt.Analyzer):

    params = (
        ('scheme', combine_custom_conf('scheme', Tradimo())),
        ('style', combine_custom_conf('style', 'bar')),
        ('lookback', 23),
        ('address', combine_custom_conf('address', 'localhost')),
        ('port', combine_custom_conf('port', 8999)),
        ('title', combine_custom_conf('title', None)),
    )

    def __init__(self, **kwargs):
        title = self.p.title
        if title is None:
            title = 'Live %s' % type(self.strategy).__name__
        if not 'autostart' in kwargs:
            kwargs['autostart'] = True
        self._webapp = Webapp(
            title,
            'basic.html.j2',
            self.p.scheme,
            self._app_cb_build_root_model,
            on_session_destroyed=self._on_session_destroyed,
            autostart=kwargs['autostart'],
            address=self.p.address,
            port=self.p.port)
        self._lock = Lock()
        self._clients = {}
        self._app_kwargs = kwargs

    def _create_app(self):
        return BacktraderBokeh(
            style=self.p.style,
            scheme=self.p.scheme,
            **self._app_kwargs)

    def _on_session_destroyed(self, session_context):
        with self._lock:
            self._clients[session_context.id].stop()
            del self._clients[session_context.id]

    def _t_server(self):
        asyncio.set_event_loop(asyncio.new_event_loop())
        loop = tornado.ioloop.IOLoop.current()
        self._webapp.start(loop)

    def _app_cb_build_root_model(self, doc):
        client = LiveClient(doc,
                            self._create_app(),
                            self.strategy,
                            self.p.lookback)
        with self._lock:
            self._clients[doc.session_context.id] = client
        return client.model

    def start(self):
        '''
        Start from backtrader
        '''
        _logger.debug('Starting PlotListener...')
        t = threading.Thread(target=self._t_server)
        t.daemon = True
        t.start()

    def stop(self):
        '''
        Stop from backtrader
        '''
        _logger.debug('Stopping PlotListener...')
        with self._lock:
            for c in self._clients.values():
                c.stop()

    def next(self):
        '''
        Next from backtrader, new data arrives
        '''
        with self._lock:
            for c in self._clients.values():
                c.next()



#!/usr/bin/env python
# -*- coding: utf-8; py-indent-offset:4 -*-
#
# Author: Metaer @ 2022/5/31  
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


from ..helper.datatable import ColummDataType


def datatable(self):
    cols1 = [['ticker', ColummDataType.STRING], ['chng%', ColummDataType.PERCENTAGE], ['cumpnl', ColummDataType.FLOAT], ['datein', ColummDataType.STRING], ['dateout', ColummDataType.STRING], ['dir', ColummDataType.STRING], ['mae%', ColummDataType.PERCENTAGE], ['mfe%', ColummDataType.PERCENTAGE], ['nbars', ColummDataType.INT], ['pnl%', ColummDataType.PERCENTAGE], ['pnl', ColummDataType.FLOAT], ['pnl/bar', ColummDataType.FLOAT], ['pricein', ColummDataType.FLOAT], ['priceout', ColummDataType.FLOAT], ['ref', ColummDataType.INT], ['size', ColummDataType.INT],  ['value', ColummDataType.FLOAT]]
    cols = []
    a = self.get_analysis()
    for one in a:
        cols1[0].append(one['ticker'])
        cols1[1].append(one['chng%']/100)
        cols1[2].append(one['cumpnl'])
        cols1[3].append(one['datein'])
        cols1[4].append(one['dateout'])
        cols1[5].append(one['dir'])
        cols1[6].append(one['mae%']/100)
        cols1[7].append(one['mfe%']/100)
        cols1[8].append(one['nbars'])
        cols1[9].append(one['pnl%']/100)
        cols1[10].append(one['pnl'])
        cols1[11].append(one['pnl/bar'])
        cols1[12].append(one['pricein'])
        cols1[13].append(one['priceout'])
        cols1[14].append(one['ref'])
        cols1[15].append(one['size'])
        cols1[16].append(one['value'])
        cols.append(cols1)
    return 'Trade List', cols



from ..helper.datatable import ColummDataType


def datatable(self):
    cols1 = [['DateTime', ColummDataType.DATETIME], ['Value', ColummDataType.FLOAT]]

    a = self.get_analysis()

    for k, v in a.items():
        cols1[0].append(k)
        cols1[1].append(v)

    return 'Calmar', [cols1]



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from collections import defaultdict
from functools import partial

from pandas import DataFrame

from bokeh.models import ColumnDataSource
from bokeh.layouts import column
from bokeh.models.widgets import DataTable, TableColumn, \
    NumberFormatter, StringFormatter

from .webapp import Webapp


class OptBrowser:
    def __init__(self, app, optresults, usercolumns=None,
                 num_result_limit=None, sortcolumn=None,
                 sortasc=True, address='localhost',port=80, autostart=True):
        self._usercolumns = {} if usercolumns is None else usercolumns
        self._num_result_limit = num_result_limit
        self._app = app
        self._sortcolumn = sortcolumn
        self._sortasc = sortasc
        self._optresults = optresults
        self._address = app.p.address
        self._port = app.p.port
        self._autostart = app.p.autostart
    
    # Metaer: self._app.params.scheme is default scheme settings. fix it
    def start(self, ioloop=None):
        webapp = Webapp(
            'Backtrader Optimization Result',
            'basic.html.j2',
            self._app.scheme,
            self.build_optresult_model,
            port=self._port,
            address=self._address,
            autostart=self._autostart)
        webapp.start(ioloop)

    def _build_optresult_selector(self, optresults):
        # 1. build a dict with all params and all user columns
        data_dict = defaultdict(list)
        for optres in optresults:
            for param_name, _ in optres[0].params._getitems():
                param_val = optres[0].params._get(param_name)
                data_dict[param_name].append(param_val)

            for usercol_label, usercol_fnc in self._usercolumns.items():
                data_dict[usercol_label].append(usercol_fnc(optres))

        # 2. build a pandas DataFrame
        df = DataFrame(data_dict)

        # 3. now sort and limit result
        if self._sortcolumn is not None:
            df = df.sort_values(by=[self._sortcolumn], ascending=self._sortasc)

        if self._num_result_limit is not None:
            df = df.head(self._num_result_limit)

        # 4. build column info for Bokeh table
        tab_columns = []

        for colname in data_dict.keys():
            formatter = NumberFormatter(format='0.000')

            if (len(data_dict[colname]) > 0
                    and isinstance(data_dict[colname][0], int)):
                formatter = StringFormatter()

            tab_columns.append(
                TableColumn(
                    field=colname,
                    title=f'{colname}',
                    sortable=False,
                    formatter=formatter))

        cds = ColumnDataSource(df)
        selector = DataTable(
            source=cds,
            columns=tab_columns,
            height=150,  # fixed height for selector
            width=0,  # set width to 0 so there is no min_width
            sizing_mode='stretch_width',
            fit_columns=True)
        return selector, cds

    def build_optresult_model(self, _=None):
        '''
        Generates and returns an interactive model for an OptResult
        or an OrderedOptResult
        '''

        def _get_model(selector_cds, idx: int):
            selector_cds.selected.indices = [idx]
            selected = selector_cds.data['index'][idx]
            return self._app.plot_optmodel(
                self._optresults[selected][0])

        # we have list of results, each result contains the result for
        # one strategy. we don't support having more than one strategy!
        if len(self._optresults) > 0 and len(self._optresults[0]) > 1:
            raise RuntimeError(
                'You passed on optimization result based on more than'
                + ' one strategy which is not supported!')

        selector, selector_cds = self._build_optresult_selector(
            self._optresults)

        # show the first result in list as default
        model = column(
            [selector, _get_model(selector_cds, 0)],
            sizing_mode='stretch_width')
        model.background = self._app.params.scheme.background_fill

        def update(selector_cds, name, old, new):
            if len(new) == 0:
                return
            stratidx = new[0]
            model.children[-1] = _get_model(
                selector_cds, stratidx)

        selector_cds.selected.on_change(
            'indices', partial(update, selector_cds))

        return model



from ..helper.datatable import ColummDataType


def datatable(self):
    cols = [['', ColummDataType.STRING], ['Value', ColummDataType.FLOAT]]
    cols[0].append('Sharpe-Ratio')

    a = self.get_analysis()
    if len(a):
        cols[1].append(a['sharperatio'])
    else:
        cols[1].append('')
    return 'Sharpe-Ratio', [cols]



from collections import defaultdict

import numpy as np
import pandas as pd

from bokeh.models import ColumnDataSource


class CDSObject:

    '''
    Base class for FigurePage and Figure with ColumnDataSource support
    also alows to create custom columns which are not available in
    provided data.
    It will create data for stream, patch and set up the columns
    in ColumnDataSource

    It is using index and datetime columns as special cases:

    -index is added, so stream has also the real index of the row
     without it, the index would be resetted in ColumnDataSource
    -datetime is added only if there are any rows to prevent gaps
     in data, so this column should only be set in cds_cols if all
     values needs to be added

    This special cases will be available in every row
    '''

    def __init__(self, cols=[]):
        self._cds_cols = []
        self._cds_cols_default = cols
        self._cds = ColumnDataSource()
        self.set_cds_col(cols)

    @property
    def cds(self):
        '''
        Property for ColumnDataSource
        '''
        return self._cds

    @property
    def cds_cols(self):
        '''
        Property for Columns in ColumnDataSource
        '''
        return self._cds_cols

    def _get_cds_cols(self):
        '''
        Returns all set columns
        2 lissts will be returned:
        - columns: columns from data source
        - additional: additional data sources which should be
          created from data source
        '''
        columns = []
        additional = []
        for c in self._cds_cols:
            if isinstance(c, str):
                columns.append(c)
            else:
                additional.append(c)
        return columns, additional

    def _create_cds_col_from_df(self, op, df):
        '''
        Creates a column from DataFrame
        op - tuple: [0] - name of column
                    [1] - source column
                    [2] - other column or value
                    [3] - op method (callable with 2 params: a, b)
        '''
        a = np.array(df[op[1]])
        if isinstance(op[2], str):
            b = np.array(df[op[2]])
        else:
            b = np.full(df.shape[0], op[2])
        arr = op[3](a, b)
        return arr

    def _create_cds_col_from_series(self, op, series):
        '''
        Creates a column from Series
        '''
        arr = self._create_cds_col_from_df(op, pd.DataFrame([series]))
        return arr[0]

    def set_cds_col(self, col):
        '''
        Sets ColumnDataSource columns to use
        allowed column types are string, tuple
        col can contain multiple columns in a list
        tuples will be used to create a new column from
        existing columns
        '''
        if not isinstance(col, list):
            col = [col]
        for c in col:
            if isinstance(c, str):
                if c not in self._cds_cols:
                    self._cds_cols.append(c)
            elif isinstance(c, tuple) and len(c) == 4:
                self._cds_cols.append(c)
            else:
                raise Exception("Unsupported col provided")

    def set_cds_columns_from_df(self, df, dropna=True):
        '''
        Sets the ColumnDataSource columns based on the given DataFrame using
        the given columns. Only the given columns will be added, all will be
        added if columns=None
        '''
        columns, additional = self._get_cds_cols()
        if not len(columns) > 0:
            columns = list(df.columns)
        try:
            c_df = df.loc[:, columns]
        except Exception:
            return None
        # remove empty rows
        if dropna:
            c_df = c_df.dropna(how='all')
        # use text NaN for nan values
        c_df.fillna('NaN')
        # ensure df contains corresponding datetime entries
        c_df['datetime'] = df.loc[
            c_df.index, 'datetime'].to_numpy(dtype=np.datetime64)

        # ensure df contains index
        c_df['index'] = df.loc[
            c_df.index, 'index'].to_numpy(dtype=np.int64)

        # add tooltip cds
        if (hasattr(self,'_hover') and hasattr(self._hover,'tooltips')):
            for tip in self._hover.tooltips:
                if tip[0] != 'Time':
                    which_tip = tip[1].split("{")[0][1:]
                    if not (which_tip in self.cds_cols):
                        c_df[which_tip] = df.loc[
                            c_df.index, which_tip]

        # add additional columns
        for a in additional:
            col = self._create_cds_col_from_df(a, c_df)
            c_df[a[0]] = col
        
        # set cds
        for c in c_df.columns:
            if c in self._cds.column_names:
                self._cds.remove(c)
            self._cds.add(np.array(c_df[c]), c)

    def get_cds_streamdata_from_df(self, df):
        '''
        Creates stream data from a pandas DataFrame
        '''
        columns, additional = self._get_cds_cols()
        if not len(columns) > 0:
            columns = list(df.columns)
        try:
            c_df = df.loc[:, columns]
        except Exception:
            return {}
        # use text NaN for nan values
        c_df.fillna('NaN')
        # ensure c_df contains datetime
        c_df['datetime'] = df.loc[c_df.index, 'datetime']
        # add additional columns
        for a in additional:
            col = self._create_cds_col_from_df(a, c_df)
            c_df[a[0]] = col
        res = ColumnDataSource.from_df(c_df)
        return res

    def get_cds_patchdata_from_series(self, series, fill_nan=[]):
        '''
        Creates patch data from a pandas Series
        '''
        p_data = defaultdict(list)
        s_data = defaultdict(list)
        columns, additional = self._get_cds_cols()
        idx_map = {d: idx for idx, d in enumerate(self._cds.data['index'])}
        # get the index in cds for series index
        if series['index'] in idx_map:
            idx = idx_map[series['index']]
        else:
            idx = False
        # create patch or stream data based on given series
        if idx is not False:
            # ensure datetime is checked for changes
            if 'datetime' not in columns:
                columns.append('datetime')
            for c in columns:
                val = series[c]
                if c == 'datetime':
                    val = val.to_numpy()
                cds_val = self._cds.data[c][idx]
                if c in fill_nan or cds_val != val:
                    if val != val:
                        val = 'NaN'
                    p_data[c].append((idx, val))
            for a in additional:
                c = a[0]
                cds_val = self._cds.data[c][idx]
                val = self._create_cds_col_from_series(a, series)
                if c in fill_nan or cds_val != val:
                    if val != val:
                        val = 'NaN'
                    p_data[c].append((idx, val))
        else:
            # add all columns to stream result. This may be needed if a value
            # was nan and therefore not added before
            s_data = self.get_cds_streamdata_from_df(pd.DataFrame([series]))
        return p_data, s_data

    def cds_reset(self):
        '''
        Resets the ColumnDataSource and other config to default
        '''
        self._cds = ColumnDataSource()
        self._cds_cols = []
        self.set_cds_col(self._cds_cols_default)



from ..helper.datatable import ColummDataType


def datatable(self):
    table1 = [['Name', ColummDataType.STRING], ['Value', ColummDataType.FLOAT]]

    a = self.get_analysis()

    table1[0].append('Average')
    table1[1].append(a.average)

    table1[0].append('Standard Deviation')
    table1[1].append(a.stddev)

    table1[0].append('Positive #')
    table1[1].append(a.positive)

    table1[0].append('Negative #')
    table1[1].append(a.negative)

    table1[0].append('Neutral #')
    table1[1].append(a.nochange)

    table1[0].append('Best')
    table1[1].append(a.best)

    table1[0].append('Worst')
    table1[1].append(a.worst)

    return 'Period Stats', [table1]




import backtrader as bt

from .params import get_params_str
from ..utils import get_clock_obj, get_dataname


def obj2label(obj, fullid=False):
    if isinstance(obj, bt.Strategy):
        return strategy2label(obj, fullid)
    elif isinstance(obj, bt.IndicatorBase):
        return indicator2label(obj, fullid)
    elif isinstance(obj, bt.AbstractDataBase):
        return data2label(obj, fullid)
    elif isinstance(obj, bt.ObserverBase):
        return observer2label(obj, fullid)
    elif isinstance(obj, bt.Analyzer):
        return obj.__class__.__name__
    elif isinstance(obj, bt.MultiCoupler):
        return obj2label(obj.datas[0], fullid)
    elif isinstance(obj, (bt.LinesOperation,
                          bt.LineSingle,
                          bt.LineSeriesStub)):
        return obj.__class__.__name__
    else:
        raise RuntimeError(f'Unsupported type: {obj.__class__}')


def strategy2label(strategy, params=False):
    label = strategy.__class__.__name__
    if params:
        param_labels = get_params_str(strategy.params)
        if len(param_labels) > 0:
            label += f' [{param_labels}]'
    return label


def data2label(data, fullid=False):
    if fullid:
        return f'{get_dataname(data)}-{data.__class__.__name__}'
    else:
        return get_dataname(data)


def observer2label(obs, fullid=False):
    if fullid:
        return obs.plotlabel()
    else:
        return obs.plotinfo.plotname or obs.__class__.__name__


def indicator2label(ind, fullid=False):
    if fullid:
        return ind.plotlabel()
    else:
        return ind.plotinfo.plotname or ind.__class__.__name__


def obj2data(obj):
    '''
    Returns a string listing all involved data feeds. Empty string if
    there is only a single feed in the mix
    '''
    if isinstance(obj, bt.LineActions):
        return 'Line Action'
    elif isinstance(obj, bt.AbstractDataBase):
        return obj2label(obj)
    elif isinstance(obj, bt.IndicatorBase):
        names = []
        for x in obj.datas:
            if isinstance(x, bt.AbstractDataBase):
                return obj2label(x)
            elif isinstance(x, bt.IndicatorBase):
                names.append(indicator2label(x, False))
            elif isinstance(x, bt.LineSeriesStub):
                # indicator target is one specific line of a datafeed
                # add [L] at the end
                return obj2label(get_clock_obj(x)) + ' [L]'
        if len(names) > 0:
            return ",".join(names)
    else:
        raise RuntimeError(f'Unsupported type: {obj.__class__}')
    return ''



from ..helper.datatable import ColummDataType


def datatable(self):
    cols1 = [['Name', ColummDataType.STRING], ['Value', ColummDataType.FLOAT]]

    a = self.get_analysis()
    cols1[0].append('VWR')
    cols1[1].append(a['vwr'])

    return 'Variability-Weighted Return', [cols1]



from ..helper.datatable import ColummDataType


def datatable(self):
    def gdef(obj, attr, d):
        return obj[attr] if attr in obj else d

    a = self.get_analysis()
    tables = []

    tab1 = [['', ColummDataType.STRING], ['Total', ColummDataType.INT], ['Open', ColummDataType.INT], ['Closed', ColummDataType.INT]]
    tab1[0].append('Number of Trades')
    tab1[1].append(a['total']['total'])
    if 'open' in a['total']:
        tab1[2].append(a['total']['open'])
    else:
        tab1[2].append('-')
    tab1[3].append(gdef(a['total'], 'closed', 0))
    tables.append(tab1)

    if 'streak' in a and 'pnl' in a:
        tab2 = [['Streak', ColummDataType.STRING], ['Current', ColummDataType.INT], ['Longest', ColummDataType.INT]]
        tab2[0].append('Won')
        tab2[1].append(a['streak']['won']['current'])
        tab2[2].append(a['streak']['won']['longest'])

        tab2[0].append('Lost')
        tab2[1].append(a['streak']['lost']['current'])
        tab2[2].append(a['streak']['lost']['longest'])

        tables.append(tab2)

        tab3 = [['Profit & Loss', ColummDataType.STRING], ['Total', ColummDataType.FLOAT], ['Average', ColummDataType.FLOAT]]
        tab3[0].append('Gross Profit')
        tab3[1].append(a['pnl']['gross']['total'])
        tab3[2].append(a['pnl']['gross']['average'])

        tab3[0].append('Net Profit (w/ Commissions)')
        tab3[1].append(a['pnl']['net']['total'])
        tab3[2].append(a['pnl']['net']['average'])

        tab3[0].append('Short')
        tab3[1].append(a['short']['pnl']['total'])
        tab3[2].append(a['short']['pnl']['average'])

        tab3[0].append('Long')
        tab3[1].append(a['long']['pnl']['total'])
        tab3[2].append(a['long']['pnl']['average'])

        tab3[0].append('Won / Short')
        tab3[1].append(a['short']['pnl']['won']['total'])
        tab3[2].append(a['short']['pnl']['won']['average'])

        tab3[0].append('Lost / Short')
        tab3[1].append(a['short']['pnl']['lost']['total'])
        tab3[2].append(a['short']['pnl']['lost']['average'])

        tab3[0].append('Won / Long')
        tab3[1].append(a['long']['pnl']['won']['total'])
        tab3[2].append(a['long']['pnl']['won']['average'])

        tab3[0].append('Lost / Long')
        tab3[1].append(a['long']['pnl']['lost']['total'])
        tab3[2].append(a['long']['pnl']['lost']['average'])

        tables.append(tab3)

        tab4 = [['Long', ColummDataType.STRING], ['Gross', ColummDataType.FLOAT], ['Net', ColummDataType.FLOAT]]
        tab4[0].append('Longest')
        tab4[1].append(a['streak']['won']['longest'])
        tab4[2].append(a['streak']['lost']['longest'])
        tables.append(tab4)

        tab5 = [['Trades', ColummDataType.STRING], ['Total', ColummDataType.INT], ['Won', ColummDataType.INT], ['Lost', ColummDataType.INT]]
        tab5[0].append('Long')
        tab5[1].append(a['long']['total'])
        tab5[2].append(a['long']['won'])
        tab5[3].append(a['long']['lost'])

        tab5[0].append('Short')
        tab5[1].append(a['short']['total'])
        tab5[2].append(a['short']['won'])
        tab5[3].append(a['short']['lost'])

        tab5[0].append('All')
        tab5[1].append(a['won']['total'] + a['lost']['total'])
        tab5[2].append(a['won']['total'])
        tab5[3].append(a['lost']['total'])

        tables.append(tab5)

        tab_len = [['Trade Length', ColummDataType.STRING], ['Total', ColummDataType.INT], ['Min', ColummDataType.INT], ['Max', ColummDataType.INT], ['Average', ColummDataType.FLOAT]]
        tab_len[0].append('Won')
        tab_len[1].append(a['len']['won']['total'])
        if 'min' in a['len']['won']:
            tab_len[2].append(a['len']['won']['min'])
        else:
            tab_len[2].append(float('nan'))
        tab_len[3].append(a['len']['won']['max'])
        tab_len[4].append(a['len']['won']['average'])

        tab_len[0].append('Lost')
        tab_len[1].append(a['len']['lost']['total'])
        if 'min' in a['len']['lost']:
            tab_len[2].append(a['len']['lost']['min'])
        else:
            tab_len[2].append(float('nan'))
        tab_len[3].append(a['len']['lost']['max'])
        tab_len[4].append(a['len']['lost']['average'])

        tab_len[0].append('Long')
        tab_len[1].append(a['len']['long']['total'])
        tab_len[2].append(a['len']['long']['min'])
        tab_len[3].append(a['len']['long']['max'])
        tab_len[4].append(a['len']['long']['average'])

        tab_len[0].append('Short')
        tab_len[1].append(a['len']['short']['total'])
        tab_len[2].append(a['len']['short']['min'])
        tab_len[3].append(a['len']['short']['max'])
        tab_len[4].append(a['len']['short']['average'])

        tab_len[0].append('Won / Long')
        tab_len[1].append(a['len']['long']['won']['total'])
        tab_len[2].append(a['len']['long']['won']['min'])
        tab_len[3].append(a['len']['long']['won']['max'])
        tab_len[4].append(a['len']['long']['won']['average'])

        tab_len[0].append('Won / Short')
        tab_len[1].append(a['len']['short']['won']['total'])
        tab_len[2].append(a['len']['short']['won']['min'])
        tab_len[3].append(a['len']['short']['won']['max'])
        tab_len[4].append(a['len']['short']['won']['average'])

        tab_len[0].append('Lost / Long')
        tab_len[1].append(a['len']['long']['lost']['total'])
        tab_len[2].append(a['len']['long']['lost']['min'])
        tab_len[3].append(a['len']['long']['lost']['max'])
        tab_len[4].append(a['len']['long']['lost']['average'])

        tab_len[0].append('Lost / Short')
        tab_len[1].append(a['len']['short']['lost']['total'])
        tab_len[2].append(a['len']['short']['lost']['min'])
        tab_len[3].append(a['len']['short']['lost']['max'])
        tab_len[4].append(a['len']['short']['lost']['average'])

        tables.append(tab_len)

    return 'Transaction Analyzer', tables



from .blackly import Blackly


class Tradimo(Blackly):
    def _set_params(self):
        super()._set_params()

        dark_text = '#333333'

        self.barup = '#e6550d'
        self.bardown = '#31a354'
        self.volup = '#ff9896'
        self.voldown = '#98df8a'
        self.barup_wick = self.barup
        self.bardown_wick = self.bardown
        self.barup_outline = self.barup
        self.bardown_outline = self.bardown

        self.crosshair_line_color = '#000000'

        self.legend_background_color = '#f5f5f5'
        self.legend_text_color = dark_text
        self.legend_click = 'hide'  # or 'mute'

        self.loc = '#265371'
        self.background_fill = 'white'
        self.body_background_color = 'white'
        self.border_fill = 'white'
        self.axis_line_color = '#222222'
        self.grid_line_color = '#eeeeee'
        self.tick_line_color = self.axis_line_color
        self.axis_text_color = dark_text
        self.plot_title_text_color = dark_text
        self.axis_label_text_color = dark_text

        self.table_color_even = '#404040'
        self.table_color_odd = '#333333'
        self.table_header_color = '#7a7a7a'

        self.tooltip_background_color = '#f5f5f5'
        self.tooltip_text_label_color = '#848EFF'
        self.tooltip_text_value_color = '#5c5c5c'

        self.tab_active_background_color = '#dddddd'
        self.tab_active_color = '#111111'

        self.text_color = '#222222'

        self.table_color_even = '#fefefefe'
        self.table_color_odd = '#eeeeee'
        self.table_header_color = '#cccccc'



from ..helper.datatable import ColummDataType


def datatable(self):
    cols = [['Time', ColummDataType.DATETIME],
            # ['Data ID', ColummDataType.INT],
            ['Size', ColummDataType.INT],
            ['Price', ColummDataType.FLOAT],
            ['Instrument', ColummDataType.STRING],
            ['Total Price', ColummDataType.FLOAT]
            ]

    # size, price, i, dname, -size * price
    for k, v in self.get_analysis().items():
        cols[0].append(k)
        # cols[1].append(v[0][2])
        cols[1].append(v[0][0])
        cols[2].append(v[0][1])
        cols[3].append(v[0][3])
        cols[4].append(v[0][4])

    return 'Transactions', [cols]



import numpy as np


def cds_op_gt(a, b):
    '''
    Operator for gt
    will create a new column with values
    from b if a > b else a
    '''
    res = np.where(a > b, b, a)
    return res


def cds_op_lt(a, b):
    '''
    Operator for lt
    will create a new column with values
    from b if a < b else a
    '''
    res = np.where(a < b, b, a)
    return res


def cds_op_non(a, b):
    '''
    Operator for non
    will return b as new column
    '''
    return b


def cds_op_color(a, b, color_up, color_down):
    '''
    Operator for color generation
    will return a column with colors
    To provide color values, use functools.partial.
    Example:
    partial(cds_op_color, color_up=color_up, color_down=color_down)
    '''
    c_up = np.full(len(a), color_up)
    c_down = np.full(len(a), color_down)
    res = np.where(b >= a, c_up, c_down)
    return res



import time
import logging
from enum import Enum
from threading import Thread, Lock
from tornado import gen

import pandas as pd

_logger = logging.getLogger(__name__)


class UpdateType(Enum):
    ADD = 1,
    UPDATE = 2,


class LiveDataHandler:

    '''
    Handler for live data
    '''

    def __init__(self, doc, app, figid, lookback, fill_gaps=True, timeout=1):
        # doc of client
        self._doc = doc
        # app instance
        self._app = app
        # figurepage id
        self._figid = figid
        # lookback length
        self._lookback = lookback
        # should gaps be filled
        self._fill_gaps = fill_gaps
        # timeout for thread
        self._timeout = timeout
        # figurepage
        self._figurepage = app.get_figurepage(figid)
        # thread to process new data
        self._thread = Thread(target=self._t_thread, daemon=True)
        self._lock = Lock()
        self._running = True
        self._new_data = False
        self._datastore = None
        self._last_idx = -1
        self._patches = []
        self._cb_patch = None
        self._cb_add = None
        # inital fill of datastore
        self._fill()
        # start thread
        self._thread.start()

    def _fill(self):
        '''
        Fills datastore with latest values
        '''
        df = self._app.generate_data(
            figid=self._figid,
            back=self._lookback,
            preserveidx=True,
            fill_gaps=self._fill_gaps)
        self._set_data(df)
        # init by calling set_cds_columns_from_df
        # after this, all cds will already contain data
        self._figurepage.set_cds_columns_from_df(self._datastore)

    def _set_data(self, data, idx=None):
        '''
        Replaces or appends data to datastore
        '''
        with self._lock:
            if isinstance(data, pd.DataFrame):
                self._datastore = data
                self._last_idx = -1
            elif isinstance(data, pd.Series):
                if idx is None:
                    # Metaer
                    # self._datastore = self._datastore.append(data)
                    self._datastore = pd.concat([self._datastore, data])
                    # end
                else:
                    self._datastore.loc[idx] = data
            else:
                raise Exception('Unsupported data provided')
            self._datastore = self._datastore.tail(
                self._get_data_stream_length())

    @gen.coroutine
    def _cb_push_adds(self):
        '''
        Streams new data to all ColumnDataSources
        '''

        # take all rows from datastore that were not yet streamed
        update_df = self._datastore[self._datastore['index'] > self._last_idx]
        # skip if we don't have new data
        if update_df.shape[0] == 0:
            return

        # store last index of streamed data
        self._last_idx = update_df['index'].iloc[-1]

        fp = self._figurepage
        # create stream data for figurepage
        data = fp.get_cds_streamdata_from_df(update_df)
        if data:
            _logger.debug(f'Sending stream for figurepage: {data}')
            fp.cds.stream(
                data, self._get_data_stream_length())

        # create stream df for every figure
        for f in fp.figures:
            data = f.get_cds_streamdata_from_df(update_df)
            if data:
                _logger.debug(f'Sending stream for figure: {data}')
                f.cds.stream(data, self._get_data_stream_length())

    @gen.coroutine
    def _cb_push_patches(self):
        '''
        Pushes patches to all ColumnDataSources
        '''
        # get all rows to patch
        patches = []
        while len(self._patches) > 0:
            patches.append(self._patches.pop(0))
        # skip if no patches available
        if len(patches) == 0:
            return

        for patch in patches:
            fp = self._figurepage

            # patch figurepage
            p_data, s_data = fp.get_cds_patchdata_from_series(patch)
            if len(p_data) > 0:
                _logger.debug(f'Sending patch for figurepage: {p_data}')
                fp.cds.patch(p_data)
            if len(s_data) > 0:
                _logger.debug(f'Sending stream for figurepage: {s_data}')
                fp.cds.stream(
                    s_data, self._get_data_stream_length())

            # patch all figures
            for f in fp.figures:
                # only fill with nan if not filling gaps
                if not self._fill_gaps:
                    c_fill_nan = f.fill_nan()
                else:
                    c_fill_nan = []
                # get patch data
                p_data, s_data = f.get_cds_patchdata_from_series(
                    patch, c_fill_nan)
                if len(p_data) > 0:
                    _logger.debug(f'Sending patch for figure: {p_data}')
                    f.cds.patch(p_data)
                if len(s_data) > 0:
                    _logger.debug(f'Sending stream for figure: {s_data}')
                    f.cds.stream(
                        s_data, self._get_data_stream_length())

    def _push_adds(self):
        doc = self._doc
        try:
            doc.remove_next_tick_callback(self._cb_add)
        except ValueError:
            pass
        self._cb_add = doc.add_next_tick_callback(
            self._cb_push_adds)

    def _push_patches(self):
        doc = self._doc
        try:
            doc.remove_next_tick_callback(self._cb_patch)
        except ValueError:
            pass
        self._cb_patch = doc.add_next_tick_callback(
            self._cb_push_patches)

    def _process(self, rows):
        '''
        Request to update data with given rows
        '''
        for idx, row in rows.iterrows():
            if (self._datastore.shape[0] > 0
                    and idx in self._datastore['index']):
                update_type = UpdateType.UPDATE
            else:
                update_type = UpdateType.ADD

            if update_type == UpdateType.UPDATE:
                ds_idx = self._datastore.loc[
                    self._datastore['index'] == idx].index[0]
                self._set_data(row, ds_idx)
                self._patches.append(row)
                self._push_patches()
            else:
                # append data and remove old data
                self._set_data(row)
                self._push_adds()

    def _t_thread(self):
        '''
        Thread method for datahandler
        '''
        while self._running:
            if self._new_data:
                last_idx = self.get_last_idx()
                last_avail_idx = self._app.get_last_idx(self._figid)
                if last_avail_idx - last_idx > (2 * self._lookback):
                    # if there is more new data then lookback length
                    # don't load from last index but from end of data
                    data = self._app.generate_data(
                        back=self._lookback,
                        preserveidx=True,
                        fill_gaps=self._fill_gaps)
                else:
                    # if there is just some new data (less then lookback)
                    # load from last index, so no data is skipped
                    data = self._app.generate_data(
                        start=last_idx,
                        preserveidx=True,
                        fill_gaps=self._fill_gaps)
                self._new_data = False
                self._process(data)
            time.sleep(self._timeout)

    def _get_data_stream_length(self):
        '''
        Returns the length of data stream to use
        '''
        return min(self._lookback, self._datastore.shape[0])

    def get_last_idx(self):
        '''
        Returns the last index in local datastore
        '''
        if self._datastore.shape[0] > 0:
            return self._datastore['index'].iloc[-1]
        return -1

    def set(self, df):
        '''
        Sets a new df and streams data
        '''
        self._set_data(df)
        self._push_adds()

    def update(self):
        '''
        Notifies datahandler of new data
        '''
        if self._running:
            self._new_data = True

    def stop(self):
        '''
        Stops the datahandler
        '''
        # mark as not running
        self._running = False
        # ensure no pending calls are set
        try:
            self._doc.remove_next_tick_callback(self._cb_patch)
        except ValueError:
            pass
        try:
            self._doc.remove_next_tick_callback(self._cb_add)
        except ValueError:
            pass
        # it would not really be neccessary to join this thread but doing
        # it for readability
        self._thread.join(0)



from .scheme import Scheme


class Blackly(Scheme):
    def _set_params(self):
        super()._set_params()

        self.barup = '#ff9896'
        self.bardown = '#98df8a'
        self.volup = '#ff9896'
        self.voldown = '#98df8a'
        self.barup_wick = self.barup
        self.bardown_wick = self.bardown
        self.barup_outline = self.barup
        self.bardown_outline = self.bardown

        self.crosshair_line_color = '#999999'

        self.legend_background_color = '#3C3F41'
        self.legend_text_color = 'lightgrey'

        self.background_fill = '#222222'
        self.body_background_color = '#2B2B2B'
        self.border_fill = '#3C3F41'
        self.legend_click = 'hide'  # or 'mute'
        self.axis_line_color = 'darkgrey'
        self.tick_line_color = self.axis_line_color
        self.grid_line_color = '#444444'
        self.axis_text_color = 'lightgrey'
        self.plot_title_text_color = 'darkgrey'
        self.axis_label_text_color = 'darkgrey'

        self.tab_active_background_color = '#666666'
        self.tab_active_color = '#bbbbbb'

        self.table_color_even = '#404040'
        self.table_color_odd = '#333333'
        self.table_header_color = '#707070'

        self.tooltip_background_color = '#4C4F51'
        self.tooltip_text_label_color = '#848EFF'
        self.tooltip_text_value_color = '#aaaaaa'

        self.tag_pre_background_color = '#222222'
        self.tag_pre_text_color = 'lightgrey'



import bisect

from datetime import datetime

import pandas as pd
import backtrader as bt


def get_slice_with_end(data, start, end):
    # end smaller than start will occur if arr is empty
    if end >= start:
        # get slice + value at end pos
        try:
            slice = list(data)[start:end] + [data[end]]
        except IndexError:
            slice = []
    else:
        slice = []
    return slice


class ClockGenerator:

    '''
    ClockGenerator will wrap a datetime line and returns clock values

    This class is used to generate a clock values which can be used with
    the ClockHandler.

    The clock values list is generated by calling get_clock. This will
    return an array with dates and the start and end index values
    matching the positions of the dates of the returned array.

    Note: the end value is increased by 1 so when slicing a line, the last
    item is included.
    '''

    def __init__(self, strategy, dataname=False):
        clk, tz = self._get_clock_values(strategy, dataname)
        self._clk = clk
        self._tz = tz

    def _get_clock_values(self, strategy, dataname):
        '''
        Returns clock values (clk, tz)
        '''
        if dataname is not False:
            data = strategy.getdatabyname(dataname)
            return data.datetime, data._tz
        # if no dataname provided, use first data
        return strategy.datetime, strategy.data._tz

    def _get_clock_array(self):
        '''
        Generates a list with dates by converting the float datetime value
        using the tz info
        '''
        # don't include nan in here by checking for x == x
        # this happens for live data with rows that are not filled yet
        clk_arr = [bt.num2date(x, self._tz) for x in self._clk.array if x == x]
        return clk_arr

    def _get_clock_range(self, arr, start=None, end=None, back=None):
        '''
        Returns the range (start, end)
        '''
        if start is None:
            start = 0
        elif type(start) == datetime:
            start = bisect.bisect_left(arr, start)
        if end is None:
            end = len(arr) - 1
        elif type(end) == datetime:
            end = bisect.bisect_right(arr, end) - 1
        # if back is provided, move back from end, override start
        if back:
            # prevent negative start int
            start = max(0, end - back + 1)
        return start, end

    def get_clock(self, start=None, end=None, back=None):
        '''
        Returns clock values for given start, end and back values

        If no start and no end is provided, the whole clock will be
        returned.
        If back is provided, then start will be overriden.
        '''
        arr = self._get_clock_array()
        start, end = self._get_clock_range(arr, start, end, back)
        arr = get_slice_with_end(arr, start, end)
        return arr, start, end

    def get_clock_time_at_idx(self, idx):
        return bt.num2date(self._clk.array[idx], self._tz)


class ClockHandler:

    '''
    ClockHandler will generate a slice of data

    It is using a clock generated by ClockGenerator and align the data
    to another clock.
    '''

    def __init__(self, clk, start, end):
        self.clk = clk
        self.start = start
        self.end = end

    def _get_data_from_list(self, llist, clk, fill_gaps=False):
        '''
        Generates data based on given clock
        '''
        data = []
        c_idx = 0
        sc = None
        # go through all clock values and align given data to clock
        for c in clk:
            v = float('nan')
            sc_prev = None
            idx_prev = None
            # go through all values to align starting from last index
            for sc_idx in range(c_idx, len(llist)):
                sc = self.clk[sc_idx]
                if sc < c:
                    # src clck is smaller: strat -> data_n
                    if llist[sc_idx] == llist[sc_idx]:
                        v = llist[sc_idx]
                elif sc == c:
                    # src clk equals: data_n -> data_n
                    if v != v or llist[sc_idx] == llist[sc_idx]:
                        v = llist[sc_idx]
                    c_idx = sc_idx + 1
                    break
                elif sc > c:
                    # scr clk bigger: data_m -> data_n
                    if fill_gaps or (sc_prev and sc_prev < c):
                        # fill only when clk value switched
                        if sc_prev:
                            v = llist[idx_prev]
                            c_idx = idx_prev + 1
                        else:
                            v = llist[sc_idx]
                            c_idx = sc_idx
                    break
                # remember current value if not nan
                if llist[sc_idx] == llist[sc_idx]:
                    sc_prev = sc
                    idx_prev = sc_idx
            # put value for c - clk pos from src clock
            data.append(v)

        return data

    def get_list_from_line(self, line, clkalign=None, fill_gaps=False):
        '''
        Returns a list with values from the given line values
        aligned to the given clock list
        '''
        llist = get_slice_with_end(line.array, self.start, self.end)

        if clkalign is None:
            clkalign = self.clk
        data = self._get_data_from_list(llist, clkalign, fill_gaps)
        return data

    def get_df_from_series(self, series, clkalign=None, name_prefix='',
                           skip=[], fill_gaps=False):
        '''
        Returns a DataFrame from the given LineSeries
        The column names will use the name_prefix and the line alias
        '''
        df = pd.DataFrame()
        for linealias in series.getlinealiases():
            if linealias in skip:
                continue
            line = getattr(series, linealias)
            df[name_prefix + linealias] = self.get_list_from_line(
                line, clkalign, fill_gaps)
        return df



from jinja2 import Environment, PackageLoader

import matplotlib.colors


def convert_color(color):
    '''
    if color is a float value then it is interpreted as a shade of grey
    and converted to the corresponding html color code
    '''
    try:
        val = round(float(color) * 255.0)
        hex_string = '#{0:02x}{0:02x}{0:02x}'.format(val)
        return hex_string
    except ValueError:
        return matplotlib.colors.to_hex(color)


def sanitize_source_name(name: str):
    '''
    removes illegal characters from source name to make it
    compatible with Bokeh
    '''
    forbidden_chars = ' (),.-/*:'
    for fc in forbidden_chars:
        name = name.replace(fc, '_')
    return name


def generate_stylesheet(scheme, template='basic.css.j2'):
    '''
    Generates stylesheet with values from scheme
    '''
    env = Environment(loader=PackageLoader('noba', 'templates'))
    templ = env.get_template(template)

    css = templ.render(dict(
        datatable_row_color_even=scheme.table_color_even,
        datatable_row_color_odd=scheme.table_color_odd,
        datatable_header_color=scheme.table_header_color,
        tab_active_background_color=scheme.tab_active_background_color,
        tab_active_color=scheme.tab_active_color,

        tooltip_background_color=scheme.tooltip_background_color,
        tooltip_text_color_label=scheme.tooltip_text_label_color,
        tooltip_text_color_value=scheme.tooltip_text_value_color,
        body_background_color=scheme.body_background_color,
        tag_pre_background_color=scheme.tag_pre_background_color,
        tag_pre_text_color=scheme.tag_pre_text_color,
        headline_color=scheme.plot_title_text_color,
        text_color=scheme.text_color))
    return css



from .confighandler import custom_config
import backtrader as bt
bt.custom_config = custom_config
import noba.tabs as tabs
#tabs: AnalyzerTab, MetadataTab, ConfigTab, LogTabs, SourceTab
bt.tabs = tabs

bt.tabs.LogTabs = lambda cols:type("LogTab",(bt.tabs.LogTab,),{'cols':cols})
bt.getlogger = bt.tabs.log.getlogger
bt.get_order_logger = lambda name="Order Trace": bt.getlogger(col=['Day', 'Ref', 'OrdType', 'Status', 'Size', 'Remsize', 'Alive'], name=name)


from .app import BacktraderBokeh
from .analyzers import LivePlotAnalyzer as BacktraderBokehLive
from .optbrowser import OptBrowser as BacktraderBokehOptBrowser
from .analyzers import RecorderAnalyzer
from .analyzers import TradelistAnalyzer
from .feeds import FakeFeed

import noba.schemes as schemes

# initialize analyzer tables
from .analyzer_tables import inject_datatables
inject_datatables()

from .plugin import *
bt.Bokeh = BacktraderBokeh

# Customized analyzers and feeds
bt.analyzers.Live = BacktraderBokehLive
bt.analyzers.Recorder = bt.analyzers.RecorderAnalyzer = RecorderAnalyzer
bt.analyzers.Tradelist = bt.analyzers.TradelistAnalyzer = TradelistAnalyzer
bt.feeds.FakeFeed = FakeFeed

bt.Opt = BacktraderBokehOptBrowser

#schemes: Blackly, Tradimo
bt.schemes = schemes
bt.schemes.Black = schemes.Blackly
bt.schemes.White = schemes.Tradimo


from ..helper.datatable import ColummDataType


def datatable(self):
    cols1 = [['', ColummDataType.STRING], ['Value', ColummDataType.FLOAT]]

    a = self.get_analysis()
    cols1[0].append('SystemQualityNumber')
    cols1[1].append(a['sqn'])

    cols1[0].append('Trades')
    cols1[1].append(a['trades'])

    return 'SystemQualityNumber', [cols1]



import copy
import logging
import backtrader as bt

_logger = logging.getLogger(__name__)


class RecorderAnalyzer(bt.Analyzer):
    def __init__(self):
        self.nexts = []

    @staticmethod
    def print_line_snapshot(name, snapshot):
        line = snapshot['array']
        if name == 'datetime':
            line = [bt.num2date(x) for x in line]
        _logger.debug(f"Line '{name:20}' idx: {snapshot['idx']} - lencount: {snapshot['lencount']} - {list(reversed(line))}")

    @staticmethod
    def print_next(idx, next):
        _logger.debug(f'--- Next: {next["prenext"]} - #{idx}')
        __class__.print_line_snapshot('datetime', next['strategy']['datetime'])

        for di, data in enumerate(next['datas']):
            _logger.debug(f'\t--- Data {di}')
            for k, v in data[1].items():
                __class__.print_line_snapshot(k, v)

        for oi, obs in enumerate(next['observers']):
            _logger.debug(f'\t--- Obvserver {oi}')
            for k, v in obs[1].items():
                __class__.print_line_snapshot(k, v)

    @staticmethod
    def print_nexts(nexts):
        for i, n in enumerate(nexts):
            __class__.print_next(i, n)

    @staticmethod
    def _copy_lines(data):
        lines = {}

        for lineidx in range(data.lines.size()):
            line = data.lines[lineidx]
            linealias = data.lines._getlinealias(lineidx)
            lines[linealias] = {'idx': line.idx, 'lencount': line.lencount, 'array': copy.deepcopy(line.array)}

        return lines

    def _record_data(self, strat, is_prenext=False):
        curbars = []
        for i, d in enumerate(strat.datas):
            curbars.append((d._name, self._copy_lines(d)))

        oblines = []
        for obs in strat.getobservers():
            oblines.append((obs.__class__, self._copy_lines(obs)))

        self.nexts.append({'prenext': is_prenext, 'strategy': self._copy_lines(strat), 'datas': curbars, 'observers': oblines})

        _logger.debug(f"------------------- next")
        self.print_next(len(strat), self.nexts[-1])
        _logger.debug(f"------------------- next-end")

    def next(self):
        for s in self.strategy.env.runningstrats:
            minper = s._getminperstatus()
            if minper > 0:
                continue
            self._record_data(s)



import math
import datetime
import logging

import backtrader as bt
from enum import Enum

_logger = logging.getLogger(__name__)


class FakeFeed(bt.DataBase):
    class State(Enum):
        BACKTEST = 0,
        BACKFILL = 1,
        LIVE = 2,

    params = (
        ('starting_value', 200),
        ('tick_interval', datetime.timedelta(seconds=25)),
        ('start_delay', 0),
        ('run_duration', datetime.timedelta(seconds=30)),  # only used when not backtest mode
        ('num_gen_bars', 10),  # number of bars to generate in backtest or backfill mode
        ('live', True),
    )

    def __init__(self):
        super(FakeFeed, self).__init__()

        self._last_delivered = None

        self._cur_value = None
        self._current_comp = 0
        self._num_bars_delivered = 0
        self._compression_in_effect = None
        self._tmoffset = datetime.timedelta(seconds=-0.5)  # configure offset cause we are sending slightly delayed ticked data (of course!)
        self._start_ts = None  # time of the first call to _load to obey start_delay

    def start(self):
        super(FakeFeed, self).start()

        self._start_ts = datetime.datetime.now()
        self._cur_value = self.p.starting_value

    def islive(self):
        return self.p.live

    def _update_line(self, dt, value):
        _logger.debug(f"{self._name} - Updating line - Bar Time: {dt} - Value: {value}")

        self.lines.datetime[0] = bt.date2num(dt)
        self.lines.volume[0] = 0.0
        self.lines.openinterest[0] = 0.0

        # Put the prices into the bar
        if math.isnan(self.lines.open[0]):
            self.lines.open[0] = value
        if math.isnan(self.lines.high[0]) or value > self.lines.high[0]:
            self.lines.high[0] = value
        if math.isnan(self.lines.low[0]) or value < self.lines.low[0]:
            self.lines.low[0] = value
        self.lines.close[0] = value
        self.lines.volume[0] = 0.0
        self.lines.openinterest[0] = 0.0

    def _update_bar(self, dt, vopen, vlow, vhigh, vclose):
        _logger.debug(f"{self._name} - Updating bar - Bar Time: {dt} - Value: {vclose}")

        self.lines.datetime[0] = bt.date2num(dt)
        self.lines.volume[0] = 0.0
        self.lines.openinterest[0] = 0.0

        # Put the prices into the bar
        self.lines.open[0] = vopen
        self.lines.high[0] = vhigh
        self.lines.low[0] = vlow
        self.lines.close[0] = vclose
        self.lines.volume[0] = 0.0
        self.lines.openinterest[0] = 0.0

    def _load(self):
        now = datetime.datetime.now()
        if now - self._start_ts < datetime.timedelta(seconds=self.p.start_delay):
            return None

        bars_done = self._num_bars_delivered >= self.p.num_gen_bars

        if self.p.live:
            if now - self._start_ts > self.p.run_duration:
                return False
        else:
            if bars_done:
                return False

        if self.p.live:
            if bars_done:
                return self._load_live(now)
            else:
                return self._load_bar(now, True)
        else:
            return self._load_bar(now)

    def _load_bar(self, now, backfill=False):
        tf, comp = (self.p.timeframe, self.p.compression) if not backfill else (self._timeframe, self._compression)
        if tf == bt.TimeFrame.Ticks:
            delta = self.p.tick_interval * comp
        elif tf == bt.TimeFrame.Seconds:
            delta = datetime.timedelta(seconds=comp)
        elif tf == bt.TimeFrame.Minutes:
            delta = datetime.timedelta(minutes=comp)
        elif tf == bt.TimeFrame.Days:
            delta = datetime.timedelta(days=comp)
        else:
            raise RuntimeError(f"{self._name} - Unsupported timeframe: {self.p.timeframe}")

        if self._last_delivered is None:
            if backfill:
                self._last_delivered = self._time_floored(now - delta * self.p.num_gen_bars, tf, comp)  # go back one bar too far since we add one instantly
            else:
                self._last_delivered = self._time_floored(now, tf)

        self._last_delivered += delta

        _logger.debug(f"{self._name} - Loading bar: {self._last_delivered}")

        if backfill:
            self._update_bar(self._last_delivered, self._cur_value, self._cur_value, self._cur_value + comp, self._cur_value + comp)
            self._cur_value += comp
        else:
            self._update_line(self._last_delivered, self._cur_value)
            self._cur_value += 1

        self._num_bars_delivered += 1
        return True

    @staticmethod
    def _time_floored(now, timeframe, comp=1):
        t = now
        if timeframe in [bt.TimeFrame.Seconds, bt.TimeFrame.Ticks]:
            t -= datetime.timedelta(seconds=t.second % comp,
                                    microseconds=t.microsecond)
        elif timeframe == bt.TimeFrame.Minutes:
            t -= datetime.timedelta(minutes=t.minute % comp,
                                    seconds=t.second,
                                    microseconds=t.microsecond)
        elif timeframe == bt.TimeFrame.Days:
            if comp != 1:
                raise Exception('For timeframe days only compression of 1 is supported.')
            t -= datetime.timedelta(hours=t.hour,
                                    minutes=t.minute,
                                    seconds=t.second,
                                    microseconds=t.microsecond)
        else:
            raise Exception(f'TimeFrame {timeframe} not supported')
        return t

    def _load_live(self, now):
        tf = self.p.timeframe

        comp = self.p.compression

        if self._last_delivered is None:
            # first run, fill last_delivered
            self._last_delivered = self._time_floored(now, tf)

        if tf == bt.TimeFrame.Ticks:
            if now - self._last_delivered < self.p.tick_interval:
                return None
            _logger.debug(f"{self._name} - Delivering - now: {now} - lastDel: {self._last_delivered}")
            self._last_delivered += self.p.tick_interval
        else:
            if tf == bt.TimeFrame.Minutes:
                if now.minute == self._last_delivered.minute:
                    return None
                self._last_delivered += datetime.timedelta(minutes=1)
            elif tf == bt.TimeFrame.Days:
                if now.day == self._last_delivered.day:
                    return None
                self._last_delivered += datetime.timedelta(days=1)

        self._current_comp += 1

        if self._current_comp == comp:  # do not use self._compression as it is modified by resampler already
            self._current_comp = 0

            self._update_line(self._last_delivered, self._cur_value)
            self._cur_value += 1
            _logger.debug(f"{self._name} - Tick delivered: {self._last_delivered}")
            return True
        else:
            return None




#!/usr/bin/env python
# -*- coding: utf-8; py-indent-offset:4 -*-
#
# Author: Metaer @ 2022/5/31  
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import backtrader as bt
class TradelistAnalyzer(bt.Analyzer):
    def get_analysis(self):
        return self.trades
    def __init__(self):
        self.trades = []
        self.cumprofit = 0.0
    def notify_trade(self, trade):
        if trade.isclosed:
            brokervalue = self.strategy.broker.getvalue()
            dir = 'short'
            if trade.history[0].event.size > 0: dir = 'long'
            pricein = trade.history[len(trade.history)-1].status.price
            priceout = trade.history[len(trade.history)-1].event.price
            datein = bt.num2date(trade.history[0].status.dt)
            dateout = bt.num2date(trade.history[len(trade.history)-1].status.dt)
            if trade.data._timeframe >= bt.TimeFrame.Days:
                datein = datein.date()
                dateout = dateout.date()

            pcntchange = 100 * priceout / pricein - 100
            pnl = trade.history[len(trade.history)-1].status.pnlcomm
            pnlpcnt = 100 * pnl / brokervalue
            barlen = trade.history[len(trade.history)-1].status.barlen
            pbar = pnl / barlen
            self.cumprofit += pnl

            size = value = 0.0
            for record in trade.history:
                if abs(size) < abs(record.status.size):
                    size = record.status.size
                    value = record.status.value

            highest_in_trade = max(trade.data.high.get(ago=0, size=barlen+1))
            lowest_in_trade = min(trade.data.low.get(ago=0, size=barlen+1))
            hp = 100 * (highest_in_trade - pricein) / pricein
            lp = 100 * (lowest_in_trade - pricein) / pricein
            if dir == 'long':
                mfe = hp
                mae = lp
            if dir == 'short':
                mfe = -lp
                mae = -hp

            self.trades.append({'ref': trade.ref, 'ticker': trade.data._name, 'dir': dir,
                 'datein': datein, 'pricein': pricein, 'dateout': dateout, 'priceout': priceout,
                 'chng%': round(pcntchange, 2), 'pnl': pnl, 'pnl%': round(pnlpcnt, 2),
                 'size': size, 'value': value, 'cumpnl': self.cumprofit,
                 'nbars': barlen, 'pnl/bar': round(pbar, 2),
                 'mfe%': round(mfe, 2), 'mae%': round(mae, 2)})



'''
Marker definition used to generate markers in bokeh using matplotlib notation
'''

_mrk_fncs = {
    # '.'	m00	point
    '.': ('dot', ['color'], {'size': 8}, {}),
    # ','	m01	pixel
    ',': ('dot', ['color'], {'size': 8}, {}),
    # 'o'	m02	circle
    'o': ('circle', ['color', 'size'], {}, {}),
    # 'v'	m03	triangle_down
    'v': ('triangle', ['color', 'size'],
          {'angle': 180, 'angle_units': 'deg'}, {}),
    # '^'	m04	triangle_up
    '^': ('triangle', ['color', 'size'], {}, {}),
    # '<'	m05	triangle_left
    '<': ('triangle', ['color', 'size'],
          {'angle': -90, 'angle_units': 'deg'}, {}),
    # '>'	m06	triangle_right
    '>': ('triangle', ['color', 'size'],
          {'angle': 90, 'angle_units': 'deg'}, {}),
    # '1'	m07	tri_down
    '1': ('y', ['color', 'size'], {}, {}),
    # '2'	m08	tri_up
    '2': ('y', ['color', 'size'],
          {'angle': 180, 'angle_units': 'deg'}, {}),
    # '3'	m09	tri_left
    '3': ('y', ['color', 'size'],
          {'angle': -90, 'angle_units': 'deg'}, {}),
    # '4'	m10	tri_right
    '4': ('y', ['color', 'size'],
          {'angle': 90, 'angle_units': 'deg'}, {}),
    # '8'	m11	octagon
    '8': ('star', ['color', 'size'], {}, {}),
    # 's'	m12	square
    's': ('square', ['color', 'size'], {}, {}),
    # 'p'	m13	pentagon
    'p': ('star_dot', ['color', 'size'], {}, {}),
    # 'P'	m23	plus(filled)
    'P': ('plus', ['color', 'size'], {}, {'size': -3}),
    # '*'	m14	star
    '*': ('asterisk', ['color', 'size'], {}, {}),
    # 'h'	m15	hexagon1
    'h': ('hex', ['color', 'size'], {}, {}),
    # 'H'	m16	hexagon2
    'H': ('hex', ['color', 'size'],
          {'angle': 45, 'angle_units': 'deg'}, {}),
    # '+'	m17	plus
    '+': ('plus', ['color', 'size'], {}, {}),
    # 'x'	m18	x
    'x': ('x', ['color', 'size'], {}, {}),
    # 'X'	m24	x(filled)
    'X': ('x', ['color', 'size'], {}, {'size': -3}),
    # 'D'	m19	diamond
    'D': ('diamond_cross', ['color', 'size'], {}, {}),
    # 'd'	m20	thin_diamond
    'd': ('diamond', ['color', 'size'], {}, {}),
    # '|'	m21	vline
    '|': ('vbar', ['color'], {}, {}),
    # '_'	m22	hline
    '_': ('hbar', ['color'], {}, {}),
    # 0 (TICKLEFT)	m25	tickleft
    0: ('triangle', ['color', 'size'],
        {'angle': -90, 'angle_units': 'deg'}, {'size': -3}),
    # 1 (TICKRIGHT)	m26	tickright
    1: ('triangle', ['color', 'size'],
        {'angle': 90, 'angle_units': 'deg'}, {'size': -3}),
    # 2 (TICKUP)	m27	tickup
    2: ('triangle', ['color', 'size'], {}, {'size': -3}),
    # 3 (TICKDOWN)	m28	tickdown
    3: ('triangle', ['color', 'size'],
        {'angle': 180, 'angle_units': 'deg'}, {'size': -3}),
    # 4 (CARETLEFT)	m29	caretleft
    4: ('triangle', ['fill_color', 'color', 'size'],
        {'angle': -90, 'angle_units': 'deg'}, {}),
    # 5 (CARETRIGHT)	m30	caretright
    5: ('triangle', ['fill_color', 'color', 'size'],
        {'angle': 90, 'angle_units': 'deg'}, {}),
    # 6 (CARETUP)	m31	caretup
    6: ('triangle', ['fill_color', 'color', 'size'], {}, {}),
    # 7 (CARETDOWN)	m32	caretdown
    7: ('triangle', ['fill_color', 'color', 'size'],
        {'angle': 180, 'angle_units': 'deg'}, {}),
    # 8 (CARETLEFTBASE)	m33	caretleft(centered at base)
    8: ('triangle', ['fill_color', 'color', 'size'],
        {'angle': -90, 'angle_units': 'deg'}, {'x': 0.25}),
    # 9 (CARETRIGHTBASE)	m34	caretright(centered at base)
    9: ('triangle', ['fill_color', 'color', 'size'],
        {'angle': 90, 'angle_units': 'deg'}, {'x': -0.25}),
    # 10 (CARETUPBASE)	m35	caretup(centered at base)
    10: ('triangle', ['fill_color', 'color', 'size'],
         {}, {'y': -0.25}),
    # 11 (CARETDOWNBASE)	m36	caretdown(centered at base)
    11: ('triangle', ['fill_color', 'color', 'size'],
         {'angle': 180, 'angle_units': 'deg'}, {'y': 0.25}),
    # 'None', ' ' or ''	 	nothing
    '': ('text', ['text_color', 'text_font_size', 'text'],
         {}, {}),
    ' ': ('text', ['text_color', 'text_font_size', 'text'],
          {}, {}),
    # '$...$' text
    '$': ('text', ['text_color', 'text_font_size', 'text'],
          {}, {}),
}


def get_marker_info(marker):
    fnc_name, attrs, vals, updates = None, list(), dict(), dict()
    if isinstance(marker, (int, float)):
        fnc_name, attrs, vals, updates = _mrk_fncs[int(marker)]
    elif isinstance(marker, str):
        if not len(marker):
            # empty string
            fnc_name, attrs, vals, updates = _mrk_fncs[str(marker)]
        else:
            fnc_name, attrs, vals, updates = _mrk_fncs[str(marker)[0]]
    else:
        raise Exception(
            f'unsupported marker type {type(marker)} for {marker}')
    return fnc_name, attrs, vals, updates



import backtrader
import logging
from .drawdown import datatable as drawdown
from .sharperatio import datatable as sharperatio
from .tradeanalyzers import datatable as tradeanalyzer
from .transactions import datatable as transactions
from .calmar import datatable as calmar
from .annualreturn import datatable as annualreturn
from .leverage import datatable as leverage
from .vwr import datatable as vwr
from .timereturn import datatable as timereturn
from .sqn import datatable as sqn
from .tradelist import datatable as tradelist
from ..analyzers import TradelistAnalyzer as TradelistAnalyzer

_logger = logging.getLogger(__name__)
_DATATABLE_FNC_NAME = 'get_analysis_table'


def inject_datatables():
    '''Injects function 'get_analysis_table' to some well-known Analyzer classes.'''
    _atables = {
        backtrader.analyzers.sharpe.SharpeRatio: sharperatio,
        backtrader.analyzers.DrawDown: drawdown,
        backtrader.analyzers.TradeAnalyzer: tradeanalyzer,
        backtrader.analyzers.Transactions: transactions,
        backtrader.analyzers.Calmar: calmar,
        backtrader.analyzers.AnnualReturn: annualreturn,
        backtrader.analyzers.GrossLeverage: leverage,
        backtrader.analyzers.VariabilityWeightedReturn: vwr,
        backtrader.analyzers.TimeReturn: timereturn,
        backtrader.analyzers.SQN: sqn,
        TradelistAnalyzer: tradelist,
    }

    for cls, labdict in _atables.items():
        curlab = getattr(cls, _DATATABLE_FNC_NAME, None)
        if curlab is not None:
            _logger.warning(f"Analyzer class '{cls.__name__}' already contains a function 'get_rets_table'. Not overriding.")
            continue
        setattr(cls, _DATATABLE_FNC_NAME, labdict)



#!/usr/bin/env python
# -*- coding: utf-8; py-indent-offset:4 -*-
#
# Author: Metaer @ 2022/5/11  
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import backtrader as bt

defaultplot = {
    'subplot': False
}

class ExtraLineMeta(bt.indicator.MetaIndicator):
    def __new__(cls, name, bases, d):
        return super(ExtraLineMeta,cls).__new__(cls, name, bases, d)
    def __init__(cls, name, bases, d):
        pass



import setuptools
from distutils.util import convert_path

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

main_ns = {}
ver_path = convert_path('noba/version.py')
with open(ver_path, encoding='utf-8') as ver_file:
    exec(ver_file.read(), main_ns)

setuptools.setup(
    name='noba',
    version=main_ns['__version__'],
    description='Noba is not only Backtrader',
    python_requires='>=3.6',
    author='Metaer',
    author_email='',
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='GNU General Public License v3 or later (GPLv3+)',
    url="https://github.com/iniself/noba/tree/noba_dev",
    project_urls={
        "Bug Tracker": "https://github.com/iniself/noba/issues",
        "Documentation": "https://github.com/iniself/noba/tree/noba_dev",
        "Source Code": "https://github.com/iniself/noba/tree/noba_dev",
        "Demos": "https://github.com/iniself/noba/tree/gh-pages",
    },

    # What does your project relate to?
    keywords=['trading', 'development', 'finance', 'quant', 'backtrader', 'Bokeh'],

    packages=setuptools.find_packages(),
    package_data={'noba': ['config/*.json', 'templates/*.j2', 'templates/config/*.json','templates/js/*.js']},
    classifiers = [
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Financial and Insurance Industry",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Office/Business :: Financial",
        "Topic :: Software Development"
    ],

    install_requires=[
        'backtrader',
        'bokeh>=2.4.3, <3.0.0',
        'jinja2',
        'pandas',
        'matplotlib~=3.2.2',
    ],
)



import backtrader as bt


def paramval2str(name, value):
    if value is None:  # catch None value early here!
        return str(value)
    elif name == "timeframe":
        return bt.TimeFrame.getname(value, 1)
    elif isinstance(value, float):
        return f"{value:.2f}"
    elif isinstance(value, (list, tuple)):
        vals = []
        for v in value:
            if isinstance(v, (list, tuple)):
                v = f'[{paramval2str(name, v)}]'
            vals.append(str(v))
        return ','.join(vals)
    elif isinstance(value, type):
        return value.__name__
    else:
        return str(value)


def get_nondefault_params(params: object):
    return {key: params._get(key)
            for key in params._getkeys()
            if not params.isdefault(key)}


def get_params(params):
    return {key: params._get(key) for key in params._getkeys()}


def get_params_str(params):
    user_params = get_nondefault_params(params)
    plabs = [f'{x}: {paramval2str(x, y)}' for x, y in user_params.items()]
    plabs = '/'.join(plabs)
    return plabs



from bokeh.models.widgets import Panel


class BacktraderBokehTab:

    '''
    Abstract class for tabs
    This class needs to be extended from when creating custom tabs.
    It is required to overwrite the _is_useable and _get_panel method.
    The _get_panel method needs to return a panel child and a title.
    '''

    def __init__(self, app, figurepage, client=None):
        self._app = app
        self._figurepage = figurepage
        self._client = client
        self._panel = None

    def _is_useable(self):
        raise Exception('_is_useable needs to be implemented.')

    def _get_panel(self):
        raise Exception('_get_panel needs to be implemented.')

    def is_useable(self):
        '''
        Returns if the tab is useable within the current environment
        '''
        return self._is_useable()

    def get_panel(self):
        '''
        Returns the panel to show as a tab
        '''
        child, title = self._get_panel()
        self._panel = Panel(child=child, title=title)
        return self._panel



import logging
from functools import partial

from bokeh.layouts import column, row, layout
from bokeh.models import Select, Spacer, Tabs, Button

from .datahandler import LiveDataHandler
from ..tabs import LiveTab
from ..utils import get_datanames

_logger = logging.getLogger(__name__)


class LiveClient:

    '''
    LiveClient provides live plotting functionality.
    '''

    NAV_BUTTON_WIDTH = 38

    def __init__(self, doc, app, strategy, lookback):
        self._app = app
        self._strategy = strategy
        self._refresh_fnc = None
        self._datahandler = None
        self._figurepage = None
        self._paused = False
        self._filter = ''
        # plotgroup for filter
        self.plotgroup = ''
        # amount of candles to plot
        self.lookback = lookback
        # should gaps in data be filled
        self.fill_gaps = False
        # bokeh document for client
        self.doc = doc
        # model is the root model for bokeh and will be set in baseapp
        self.model = None

        # append config tab if default tabs should be added
        if self._app.p.use_default_tabs:
            self._app.tabs.append(LiveTab)
        # set plotgroup from app params if provided
        if self._app.p.filter and self._app.p.filter['group']:
            self.plotgroup = self._app.p.filter['group']
        # create figurepage
        self._figid, self._figurepage = self._app.create_figurepage(
            self._strategy, filldata=False)
        # create model
        self.model, self._refresh_fnc = self._createmodel()
        # update model with current figurepage
        self.updatemodel()

    def _createmodel(self):

        def on_select_filter(self, a, old, new):
            _logger.debug(f'Switching filter to {new}...')
            # ensure datahandler is stopped
            self._datahandler.stop()
            self._filter = new
            self.updatemodel()
            _logger.debug('Switching filter finished')

        def on_click_nav_action(self):
            if not self._paused:
                self._pause()
            else:
                self._resume()
            refresh(self)

        def on_click_nav_prev(self, steps=1):
            self._pause()
            self._set_data_by_idx(self._datahandler.get_last_idx() - steps)
            update_nav_buttons(self)

        def on_click_nav_next(self, steps=1):
            self._pause()
            self._set_data_by_idx(self._datahandler.get_last_idx() + steps)
            update_nav_buttons(self)

        def refresh(self):
            self.doc.add_next_tick_callback(partial(update_nav_buttons, self))

        def reset_nav_buttons(self):
            btn_nav_prev.disabled = True
            btn_nav_next.disabled = True
            btn_nav_action.label = '❙❙'

        def update_nav_buttons(self):
            last_idx = self._datahandler.get_last_idx()
            last_avail_idx = self._app.get_last_idx(self._figid)

            if last_idx < self.lookback:
                btn_nav_prev.disabled = True
                btn_nav_prev_big.disabled = True
            else:
                btn_nav_prev.disabled = False
                btn_nav_prev_big.disabled = False
            if last_idx >= last_avail_idx:
                btn_nav_next.disabled = True
                btn_nav_next_big.disabled = True
            else:
                btn_nav_next.disabled = False
                btn_nav_next_big.disabled = False
            if self._paused:
                btn_nav_action.label = '▶'
            else:
                btn_nav_action.label = '❙❙'

        # filter selection
        datanames = get_datanames(self._strategy)
        options = [('', 'Strategy')]
        for d in datanames:
            options.append(('D' + d, f'Data: {d}'))
        options.append(('G', 'Plot Group'))
        self._filter = 'D' + datanames[0]
        select_filter = Select(
            value=self._filter,
            options=options)
        select_filter.on_change(
            'value',
            partial(on_select_filter, self))
        # nav
        btn_nav_prev = Button(label='❮', width=self.NAV_BUTTON_WIDTH)
        btn_nav_prev.on_click(partial(on_click_nav_prev, self))
        btn_nav_prev_big = Button(label='❮❮', width=self.NAV_BUTTON_WIDTH)
        btn_nav_prev_big.on_click(partial(on_click_nav_prev, self, 10))
        btn_nav_action = Button(label='❙❙', width=self.NAV_BUTTON_WIDTH)
        btn_nav_action.on_click(partial(on_click_nav_action, self))
        btn_nav_next = Button(label='❯', width=self.NAV_BUTTON_WIDTH)
        btn_nav_next.on_click(partial(on_click_nav_next, self))
        btn_nav_next_big = Button(label='❯❯', width=self.NAV_BUTTON_WIDTH)
        btn_nav_next_big.on_click(partial(on_click_nav_next, self, 10))
        # layout
        controls = row(
            children=[select_filter])
        nav = row(
            children=[btn_nav_prev_big,
                      btn_nav_prev,
                      btn_nav_action,
                      btn_nav_next,
                      btn_nav_next_big])
        # tabs
        tabs = Tabs(
            id='tabs',
            sizing_mode=self._app.scheme.plot_sizing_mode)
        # model
        model = layout(
            [
                # app settings, top area
                [column(controls, width_policy='min'),
                 Spacer(),
                 column(nav, width_policy='min')],
                Spacer(height=15),
                # layout for tabs
                [tabs]
            ],
            sizing_mode='stretch_width')
        return model, partial(refresh, self)

    def updatemodel(self):
        self.doc.hold()
        self._app.update_figurepage(filter=self._get_filter())
        panels = self._app.generate_model_panels()
        for t in self._app.tabs:
            tab = t(self._app, self._figurepage, self)
            if tab.is_useable():
                panels.append(tab.get_panel())

        # set all tabs (from panels, without None)
        self._get_tabs().tabs = list(filter(None.__ne__, panels))

        # create new data handler
        if self._datahandler is not None:
            self._datahandler.stop()
        self._datahandler = LiveDataHandler(
            doc=self.doc,
            app=self._app,
            figid=self._figid,
            lookback=self.lookback,
            fill_gaps=self.fill_gaps)

        # refresh model
        self._refresh_fnc()
        self.doc.unhold()

    def _get_filter(self):
        res = {}
        if self._filter.startswith('D'):
            res['dataname'] = self._filter[1:]
        elif self._filter.startswith('G'):
            res['group'] = self.plotgroup
        return res

    def _pause(self):
        self._paused = True

    def _resume(self):
        if not self._paused:
            return
        self._datahandler.update()
        self._paused = False

    def _set_data_by_idx(self, idx=None):
        # if a index is provided, ensure that index is within data range
        if idx:
            # don't allow idx to be bigger than max idx
            last_avail_idx = self._app.get_last_idx(self._figid)
            idx = min(idx, last_avail_idx)
            # don't allow idx to be smaller than lookback - 1
            idx = max(idx, self.lookback - 1)
        # create DataFrame based on last index with length of lookback
        df = self._app.generate_data(
            figid=self._figid,
            end=idx,
            back=self.lookback,
            preserveidx=True,
            fill_gaps=self.fill_gaps)
        self._datahandler.set(df)

    def _get_tabs(self):
        return self.model.select_one({'id': 'tabs'})

    def next(self):
        '''
        Request for updating data with rows
        '''
        if not self._paused:
            self._datahandler.update()
        if self._refresh_fnc:
            self._refresh_fnc()

    def stop(self):
        self._datahandler.stop()
        pass



from ..helper.datatable import ColummDataType


def datatable(self):
    cols1 = [['Feature', ColummDataType.STRING], ['Value', ColummDataType.FLOAT], ['Maximum', ColummDataType.FLOAT]]

    a = self.get_analysis()
    cols1[0].append('Length')
    cols1[1].append(a['len'])
    cols1[2].append(a['max']['len'])

    cols1[0].append('Moneydown')
    cols1[1].append(a['moneydown'])
    cols1[2].append(a['max']['moneydown'])

    cols1[0].append('Drawdown')
    cols1[1].append(a['drawdown'])
    cols1[2].append(a['max']['drawdown'])

    return 'Drawdown', [cols1]



from collections import OrderedDict
from enum import Enum

import backtrader as bt

from bokeh.models import ColumnDataSource, Paragraph, TableColumn, DataTable, \
    DateFormatter, NumberFormatter, StringFormatter

from .params import get_params_str


# the height of a single row
ROW_HEIGHT = 25


class ColummDataType(Enum):
    DATETIME = 1
    FLOAT = 2
    INT = 3
    PERCENTAGE = 4
    STRING = 5


class TableGenerator:

    '''
    Table generator for key -> value tuples
    '''

    def get_table(self, data):
        table = [['Name'], ['Value']]
        cds = ColumnDataSource()
        columns = []
        for n, v in data.items():
            table[0].append(n)
            table[1].append(v)
        for i, c in enumerate(table):
            col_name = f'col{i}'
            cds.add(c[1:], col_name)
            columns.append(TableColumn(
                field=col_name,
                title=c[0]))
        column_height = len(table[0]) * ROW_HEIGHT
        dtable = DataTable(
            source=cds,
            columns=columns,
            index_position=None,
            height=column_height,
            width=0,  # set width to 0 so there is no min_width
            sizing_mode='stretch_width',
            fit_columns=True)
        return dtable


class AnalysisTableGenerator:

    '''
    Table generator for analyzers
    '''

    def __init__(self, scheme):
        self._scheme = scheme

    @staticmethod
    def _get_table_generic(analyzer):
        '''
        Returns two columns labeled '' and 'Value'
        '''
        table = [
            ['', ColummDataType.STRING],
            ['Value', ColummDataType.STRING]]

        def add_to_table(item, baselabel=''):
            if isinstance(item, dict):
                for ak, av in item.items():
                    label = f'{baselabel} - {ak}' if len(baselabel) > 0 else ak
                    if isinstance(av, (dict, bt.AutoOrderedDict, OrderedDict)):
                        add_to_table(av, label)
                    else:
                        table[0].append(label)
                        table[1].append(av)

        add_to_table(analyzer.get_analysis())
        return analyzer.__class__.__name__, [table]

    def _get_formatter(self, ctype):
        if ctype == ColummDataType.FLOAT:
            return NumberFormatter(format=self._scheme.number_format)
        elif ctype == ColummDataType.INT:
            return NumberFormatter()
        elif ctype == ColummDataType.DATETIME:
            return DateFormatter(format='%Y-%m-%d')
        elif ctype == ColummDataType.STRING:
            return StringFormatter()
        elif ctype == ColummDataType.PERCENTAGE:
            return NumberFormatter(format='0.000 %')
        else:
            raise Exception(f'Unsupported ColumnDataType: "{ctype}"')

    def get_tables(self, analyzer):
        '''
        Return a header for this analyzer and one *or more* data tables.
        '''
        if hasattr(analyzer, 'get_analysis_table'):
            title, table_columns_list = analyzer.get_analysis_table()
        else:
            # Analyzer does not provide a table function. Use our generic one
            title, table_columns_list = __class__._get_table_generic(
                analyzer)

        # don't add empty analyzer
        if len(table_columns_list[0][0]) == 2:
            return None, None

        param_str = get_params_str(analyzer.params)
        if len(param_str) > 0:
            title += f' ({param_str})'

        elems = []
        for table_columns in table_columns_list:
            cds = ColumnDataSource()
            columns = []
            for i, c in enumerate(table_columns):
                col_name = f'col{i}'
                cds.add(c[2:], col_name)
                columns.append(TableColumn(
                    field=col_name,
                    title=c[0],
                    formatter=self._get_formatter(c[1])))
            # define height of column by multiplying count of rows
            # with ROW_HEIGHT
            column_height = len(table_columns[0]) * ROW_HEIGHT
            elems.append(DataTable(
                source=cds,
                columns=columns,
                index_position=None,
                height=column_height,
                width=0,  # set width to 0 so there is no min_width
                sizing_mode='stretch_width',
                fit_columns=True))
        return Paragraph(
            text=title,
            css_classes=['table-title']), elems
